from .provider.bitwarden_store import BitwardenStore
from .provider.keepass_store import KeepassStore
from colorama import init, Back, Style
from .get_char import GetChar
from os import system, name
import logging
import pyperclip
import argparse


def does_match(se, search=''):
    if search.lower() in se.name.lower():
        return True

    for add in se.addresses:
        if search.lower() in add.lower():
            return True

    return False


def filter_entries(store, search=''):
    return list(filter(lambda e: does_match(e, search), store.get_and_cache_simple_entries()))


def paint(search='', entries=[], idx=0):
    screen_clear()
    print('Search: ' + search)
    for x, entry in enumerate(entries):
        if x == idx:
            print(Back.MAGENTA + str(entry) + Style.RESET_ALL)

        else:
            print(str(entry))


def screen_clear():
    if name == 'nt':
        system('cls')  # windows
    else:
        system('clear')  # unix


def start_input_loop(store):
    char_list = []
    select_idx = 0
    get_char = GetChar()
    while True:
        search = ''.join(char_list)
        entries = filter_entries(store, search)
        paint(search=search, entries=entries, idx=select_idx)
        ch = get_char()
        logging.debug('char repr: ' + repr(ch))
        if ch == '\x03':  # control-c
            break
        elif ch == '\x7f':  # backspace
            select_idx = 0
            if len(char_list) > 0:
                char_list.pop(-1)
        elif ch == '\r' or ch == '\n':
            if len(entries) > select_idx:
                return entries[select_idx]
        elif ch == '\x1b':
            direction = get_char() + get_char()
            if direction == '[B' and select_idx < len(entries) - 1:
                select_idx += 1
            elif direction == '[A' and select_idx > 0:
                select_idx -= 1
        else:
            char_list += ch
            select_idx = 0
        logging.debug('char_list' + str(char_list))


def main():
    init()

    parser = argparse.ArgumentParser(description='Password Lookup')

    parser.add_argument('-f', action='store', dest='file',
                        help='KeePassFile')

    parser.add_argument('-c', action='store_true', default=False,
                        dest='clear_session',
                        help='Clear session')
    result = parser.parse_args()
    if not result.file:
        store = BitwardenStore(clear_session=result.clear_session)
    else:
        store = KeepassStore(result.file, clear_session=result.clear_session)

    filter_entries(store)
    entry = start_input_loop(store)
    if entry is not None:
        print('copied password!')
        pyperclip.copy(entry.password)


if __name__ == '__main__':
    main()



from .simple_entry import SimpleEntry
from .store_provider import StoreProvider
import libkeepass
from .session import Session
import logging
import sys
import getpass
from os import path


class KeepassStore(StoreProvider):
    def __init__(self, kee_pass_file, clear_session=False):
        self.session = Session('keepass_key', 'keepass_session')
        if clear_session:
            self.session.clear()
        self.session.load()
        self.entries = None
        self.kee_pass_file = kee_pass_file
        if not path.exists(kee_pass_file):
            logging.error('failed to open keepass file' + str(kee_pass_file))
            raise IOError('keepass file does not exist')

    def _session_is_established(self):
        if self.session.session is None or self.session.session == '':
            return False

        return True

    def load_cache(self):
        with libkeepass.open(self.kee_pass_file, password=self.session.session) as kpdb:
            self.session.write()
            self.entries = []
            for kpEntry in kpdb.obj_root.findall('.//Entry'):
                simple_entry = SimpleEntry(None, [], None)
                for s in kpEntry.findall('./String'):
                    key = s.find('./Key').text
                    val = s.find('./Value').text
                    if val is not None and val != '':
                        if key == 'URL':
                            simple_entry.addresses.append(val)
                        elif key == 'Password':
                            simple_entry.password = val
                        elif key == 'UserName':
                            simple_entry.username = val
                        elif key == 'Title':
                            simple_entry.name = val
                if simple_entry.name is not None and simple_entry.password is not None:
                    self.entries.append(simple_entry)

    def ensure_session_and_load_cache(self):
        while not self._session_is_established():
            try:
                self.session.session = getpass.getpass(
                    '\nPlease enter your password for file: ' + str(self.kee_pass_file) + ": ")

            # except CredentialsError as e:
            #     print('wrong password')
            #     self.session.session = ''
            except:
                print("Unexpected error:", sys.exc_info()[0])
                raise

    def get_and_cache_simple_entries(self):
        self.ensure_session_and_load_cache()
        self.load_cache()
        return self.entries

    def get_items(self):
        pass



import unittest
from .crypt_util import CryptUtil


class CryptUtilTest(unittest.TestCase):

    def test(self):
        cu = CryptUtil()
        key = cu.generate_key()
        print(key)
        cu.write_key_to_file('/Users/nick/PycharmProjects/password-lookup/provider/cu_test.key', key)
        key2 = cu.read_key_from_file('/Users/nick/PycharmProjects/password-lookup/provider/cu_test.key')
        self.assertTrue(key2 == key)

        msg = 'test message'
        enc_msg = cu.encrypt_str(msg, key)
        dec_msg = cu.decrypt_bytes(enc_msg, key)
        print(msg)
        print(dec_msg)
        self.assertTrue(dec_msg == msg)


if __name__ == "__main__":
    unittest.main()



from cryptography.fernet import Fernet


class CryptUtil:
    def __init__(self):
        pass

    def generate_key(self):
        return Fernet.generate_key()

    def write_key_to_file(self, file, key):
        with open(file, 'wb') as f:
            f.write(key)

    def read_key_from_file(self, file):
        with open(file, 'rb') as f:
            return f.read()

    def encrypt_str(self, message, key):
        f = Fernet(key)
        encrypted = f.encrypt(message.encode())
        return encrypted

    def decrypt_bytes(self, bytes, key):
        f = Fernet(key)
        decrypted = f.decrypt(bytes)
        return decrypted.decode()



import time

time.clock = time.process_time
import os
import sys
import logging
from .crypt_util import CryptUtil


def get_script_path():
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def get_key_filename(key_suffix):
    return (get_script_path() + "/{}.key").format(key_suffix)


def remove_key_file(key_suffix):
    if os.path.exists(get_key_filename(key_suffix)):
        os.remove(get_key_filename(key_suffix))


def get_key(key_suffix):
    if not os.path.exists(get_key_filename(key_suffix)):
        return generate_key(key_suffix)
    try:
        with open(get_key_filename(key_suffix), 'rb') as key_file:
            return key_file.readline()
    except IOError:
        return generate_key(key_suffix)


def generate_key(key_suffix):
    logging.debug('generating key')
    cu = CryptUtil()
    key = cu.generate_key()
    with open(get_key_filename(key_suffix), "wb") as key_file:
        key_file.write(key)
    return key


class Session:
    session_file = '/tmp/{}.session'

    def __init__(self, key_suffix, session_suffix):
        self.key_suffix = key_suffix
        self.session_suffix = session_suffix
        self.session = None

    def _remove_session_file(self):
        if os.path.exists(self.get_session_file()):
            os.remove(self.get_session_file())

    def clear(self):
        self.session = None
        remove_key_file(self.key_suffix)
        self._remove_session_file()

    def load(self):
        try:
            self.session = self.read()
        except:
            logging.debug("error occurred reading existing session - clearing and retrying")
            self.clear()
            self.session = self.read()

    def get_session_file(self):
        return Session.session_file.format(self.session_suffix)

    def set(self, session):
        self.session = session

    def write(self):
        with open(self.get_session_file(), "wb") as f:
            cu = CryptUtil()
            key = get_key(self.key_suffix)
            f.write(cu.encrypt_str(self.session, key))

    def read(self):
        if not os.path.exists(self.get_session_file()):
            self.session = ''
        else:
            try:
                with open(self.get_session_file(), 'rb') as f:
                    key = get_key(self.key_suffix)
                    logging.debug('decrypting with key ' + str(key))
                    cu = CryptUtil()
                    self.session = cu.decrypt_bytes(f.read(), key)
                    # Do something with the file
            except IOError:
                logging.debug("read File not accessible")
                self.session = ''
        return self.session



class GetChar:
    def __init__(self):
        try:
            self.impl = _GetCharWindows()
        except ImportError:
            self.impl = _GetCharUnix()

    def __call__(self):
        return self.impl()


class _GetCharUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetCharWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()



import unittest
from .bitwarden_store import BitwardenStore


class TestBitwardenStore(unittest.TestCase):

    def test_get_and_cache_simple_entries(self):
        bw = BitwardenStore(True)
        entries = bw.get_and_cache_simple_entries()
        self.assertTrue(entries is not None)
        for entry in entries:
            print('items: {}'.format(entry))


if __name__ == "__main__":
    unittest.main()



class SimpleEntry:
    def __init__(self, name, address, password, username=''):
        self.name = name
        self.addresses = address
        self.password = password
        self.username = username

    def __str__(self):
        return 'name: {} urls: {}'.format(self.name, ', '.join(self.addresses))



from .simple_entry import SimpleEntry
from .store_provider import StoreProvider
import subprocess
import json
import logging
import os
from .session import Session
import getpass

SESSION_KEY = 'BW_SESSION'


def get_session():
    """
    Function to return a valid Bitwarden session
    """
    # Check for an existing, user-supplied Bitwarden session
    try:
        if os.environ[SESSION_KEY]:
            logging.debug('Existing Bitwarden session found')
            return os.environ[SESSION_KEY]
    except KeyError:
        pass

    # Check if we're already logged in
    proc = subprocess.Popen(
        [
            'bw',
            'login',
            '--check',
            '--quiet'
        ]
    )
    proc.wait()

    if proc.returncode:
        logging.debug('Not logged into Bitwarden')
        operation = 'login'
        credentials = ["{}".format(input('Bitwarden user: '))]
    else:
        logging.debug('Bitwarden vault is locked')
        operation = 'unlock'
        credentials = []

    # Ask for the password
    pw = getpass.getpass('Bitwarden Vault password: ')
    credentials.append("{}".format(pw))

    logging.debug(list(filter(None, [
        'bw',
        '--raw',
        '--nointeraction',
        operation
    ] + credentials)))

    proc = subprocess.Popen(
        list(filter(None, [
            'bw',
            '--raw',
            '--nointeraction',
            operation
        ] + credentials)),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    (stdout, stderr) = proc.communicate()

    if proc.returncode:
        logging.error(stderr.decode('utf-8'))
        return None

    return stdout.decode('utf-8')


def run_bw_command(args):
    cmd = ['bw']
    cmd.extend(args)
    logging.debug('full cmd {}'.format(cmd))
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE
    )

    (stdout, _) = proc.communicate()

    if proc.returncode:
        raise RuntimeError('Unable to exec Bitwarden CLI ')

    return stdout.decode('utf-8')


def get_bw_status(session_id):
    return json.loads(run_bw_command(['status', '--session', session_id]))['status']


def build_simple_entries(items):
    entries = []
    for item in items:
        se = SimpleEntry(item['name'], [], '')
        try:
            se.addresses = list(filter(lambda x: x is not None, map(lambda entry: entry['uri'], item['login']['uris'])))
        except:
            se.addresses = []
        try:
            se.password = item['login']['password']
        except:
            continue

        entries.append(se)
    return entries


class BitwardenStore(StoreProvider):
    def __init__(self, clear_session=False):
        self.session = Session('bw_key', 'bw_session')
        if clear_session:
            self.session.clear()
        self.session.load()
        self.entries = None

    def get_and_cache_simple_entries(self):
        if self.entries:
            return self.entries
        self.entries = build_simple_entries(self.get_items())
        return self.entries

    def ensure_session(self):
        try:
            if get_bw_status(self.session.session) == 'unlocked':
                return
        except:
            pass
        session_id = get_session()
        while not session_id:
            logging.error('session id could not be established please try again')
            session_id = get_session()
        logging.debug('setting up session')
        self.session.set(session_id)
        self.session.write()

    def get_bw_version(self):
        print(run_bw_command(['--version']))

    def get_items(self):
        self.ensure_session()
        return json.loads(run_bw_command(['list', 'items', '--session', self.session.session]))



import unittest

from .keepass_store import KeepassStore


class TestBitwardenStore(unittest.TestCase):
    def test_get_and_cache_simple_entries(self):
        kps = KeepassStore('/Users/nick/PycharmProjects/password-lookup/provider/keepass.kdbx', True)
        entries = kps.get_and_cache_simple_entries()
        self.assertTrue(entries is not None)
        for entry in entries:
            print('items: {}'.format(entry))


if __name__ == "__main__":
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="passwordlookup",
    version="0.0.2",
    author="Nick Pisani",
    author_email="napisani@yahoo.com",
    description="password lookup utility for looking up keepass or bitwarden passwords via a terminal session",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/napisani/password-lookup",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'colorama==0.4.3',
        'cryptography==3.0',
        'libkeepass==0.2.0',
        'pycparser==2.20',
        'pycrypto==2.6.1',
        'pycryptodomex==3.9.8',
        'pyperclip==1.8.0',
    ]

)



import unittest
from .session import Session


class SessionTest(unittest.TestCase):

    def test(self):
        s = Session('unit_test', 'unit_test')
        s.clear()
        s.load()
        print(s.read())
        s.write()
        s.read()


if __name__ == "__main__":
    unittest.main()
